import greenfoot.*;
import java.util.List;
import java.util.ArrayList;
import javax.swing.JOptionPane;
import javax.swing.JTextArea;
/**
 *  TITLE:  CHESS: Drag & Drop<br>
 *  AUTHOR: danpost<br>
 *  VERSION: October 22, 2012<br><br>
 *  DESCRIPTION: A drag-n-drop version of chess; complete with move listing and 'UNDO' button.<br><br>
 *  WORLD CLASS: Board<br>
 *  A world in which the game of chess can be played
 */
public class Board extends World
{
    final boolean CAPTURES_ONLY = false, ALL_MOVES = true;
    // the previoius boolean values are used to determine if all or just capture moves are listed in calls to methods that list moves
    final static int[] pieceInt = { 9818, 9819, 9820, 9821, 9822, 9823, 32, 9817, 9816, 9815, 9814, 9813, 9812 };
    // the previous int values are the character values for the piece images (just like char (65) is uppercase 'A' image, char (9818) is 'Black King' image)
    String position = "rnbqkbnrpppppppp                                PPPPPPPPRNBQKBNR"; // String representation of the board position at start-up
    int moveCount = 0; // to track number of game moves played so far (and listed in 'gameMoves[]' below)
    String[] gameMoves = new String[240]; // to hold the moves played in a game
    int[] moveCastlingLost = { -1, -1, -1, -1}; // (bq, bk, wq, wk) to hold the move number when castling was no longer possible
    InfoFrame info = new InfoFrame(); // shown on the right third of the window listing the moves played in a game
    String moveList = ""; // to hold possible moves for current player
    String movers = ""; // to list the type of piece that moves for each move of the game ( (char) pieceInt: min=0; max=12;)
    String gameString = ""; // to list the game moves in condensed form

   
    /**
     * Draws a checkboard on the background of the world, adds the chessmen to the world, 
     * creates a list of possible moves, and adds the title screen to the world
     */
    public Board()
    {    
        super(1040, 600, 1);
        setPaintOrder(Panel.class, Icon.class, Piece.class);
        drawCheckerboard(); // the chessboard on the background image
        populate(); // the pieces, the information panel and the 'Undo' button
        updateMoveList(); // the initial list of possible moves for white
        showHeader(); // a title screen
        
    }
    
    // populate() adds the main objects to the world (called during world construction)
    private void populate()
    {
        addPieces();
        addObject(info, 820, 300);
        addObject(new Button("UNDO"), 821, 585);
    }
    
    // showHeader() creates and adds to the world the initial header screen panel (called during world construction)
    private void showHeader()
    {
        Panel panel = new Panel("header.png"); // the header screen object
        GreenfootImage pImg = panel.getImage(); // the image of the header screen
        // add the chess picture at top-left of header image
        GreenfootImage pic = new GreenfootImage("chess.jpg");
        pic.scale(594, 288);
        pImg.drawImage(pic, 4, 163);
        // add the notes picture at top-right of header image
        // drawing the image on a slightly larger red background gives it the frame
        pImg.setColor(Color.RED);
        pImg.fillRect(2, 451, 596, 3);
        pImg.drawImage(new GreenfootImage("notes.png"), 601, 1);
        // add the bottom frame to the header image (to be used in future update)
        GreenfootImage dImg = new GreenfootImage(1040, 146);
        dImg.setColor(Color.WHITE);
        dImg.fill(); // makes the frame opaque
        dImg.setColor(Color.RED);
        dImg.fillRect(1, 1, 1039, 144); // add the future frame
        dImg.setColor(Color.ORANGE);
        dImg.fillRect(4, 4, 1033, 138); // adds the inner background
        pImg.drawImage(dImg, 0, 454);
        addObject(panel, 520, 300); // displays the panel
    }
    
    // drawCheckerboard() draws the checkerboard onto the background of the world and draws a frame around it (called during world construction)
    private void drawCheckerboard()
    {
        GreenfootImage[] squareImage  = { new GreenfootImage(75, 75), new GreenfootImage(75, 75) }; // images for the two different colored squares
        // give each square its color
        squareImage[0].setColor(Color.LIGHT_GRAY);
        squareImage[0].fill(); 
        squareImage[1].setColor(Color.GRAY);
        squareImage[1].fill();
        // draw the squares of the checkerboard on the background of the world
        GreenfootImage bg = getBackground(); // gets the background image
        bg.clear(); // clears the background image
        for (int j = 0; j < 8; j++)
        { // for each column
            for (int i = 0; i < 8; i++)
            { // for each row
                bg.drawImage(squareImage[(i + j) % 2], i * 75, j * 75); // draw the required square
            }
        }
        // frame the checkerboard
        bg.setColor(Color.BLACK);
        bg.drawRect(0, 0, 599, 599);
    }
    
    // addPieces() puts the pieces in the world to display the current position determined by the String variable 'position'
    private void addPieces()
    {
        removeObjects(getObjects(Piece.class)); // erases any pieces currently being display
        for (int mySq = 0; mySq < position.length(); mySq++)
        { // for each character in String 'position
            String myCode = position.substring(mySq, mySq + 1); // gets the String representation of the piece on the current square
            if (!" ".equals(myCode))
            { // if square is not empty
                int myFile = mySq % 8, myRank = (mySq - myFile) / 8; // converts square number to file and rank
                addObject(new Piece(myCode), 37 + myFile * 75, 37 + myRank * 75); // displays the piece
            }
        }
    }
    
    // listMoves(String, boolean) takes a String representation of a player ("w" or "b") and a flag to determine if all or just capture moves are listed
    // and returns a concatenated String of 'legal' moves seperated by ":"
    private String listMoves(String whose, boolean listAll)
    {
        String moves = ""; // to build the list of required moves in
        // build the list of currently possible moves
        for (int sq = 0; sq < 64; sq++)
        { // for each square on the board
            int pceType = pieceAt(sq); // gets the piece on the current square
            if (pceType * (("w".equals(whose)) ? 1 : -1) > 0)
            { // if piece is of current movers ('whose') color
                moves += getMoves(pceType, sq, listAll); // adds its available moves to the move list
            }
        }
        // eliminate any illegal moves that leave the player's king under attack (in check)
        if (listAll)
        {
            String mvs = moves + ":"; // copy move list to a new variable to work off of
            moves = ""; // clear the original move list so it can be rebuild
            while (!":".equals(mvs))
            { // while moves remain in the working move list
                String mv = mvs.substring(1, mvs.substring(1).indexOf(':') + 1); // gets the current move to look at
                mvs = mvs.substring(mvs.substring(1).indexOf(':') + 1); // removes the current move from the working move list
                makeMove(mv); // executes the current move
                String king = (moveCount % 2 == 0) ? "k" : "K"; // gets a String representation of the king whose capture we are looking for
                // the next statement lists the opponent's capture moves and determines if a king capture is present or not
                boolean selfCheck = listMoves(((moveCount % 2 == 0) ? "w" : "b"), CAPTURES_ONLY).indexOf("x" + king) > -1;
                takeBack(mv); // undoes the current move
                if (!selfCheck)
                { // player's king was not in check after the move was executed
                    // if castling, must check square king passes over
                    if (mv.indexOf('<') == -1 && mv.indexOf('>') == -1) moves += ":" + mv; // not castling, add move to move list
                    else if (gameMoves[moveCount - 1].indexOf('+') == -1) // castling and not in check before move
                    {
                        // the following statement uses the condition that if the king can move one square toward to castled position,
                        // whose move is previously checked, then it is not castling through check
                        if (mv.indexOf('<') != -1 && moves.indexOf(":" + mv.substring(0, 2) + notate(getSqPos(mv.substring(0, 2)) - 1)) != -1)
                            moves += ":" + mv; // not queen-side castling through check
                        else if (mv.indexOf('>') != -1 && moves.indexOf(":" + mv.substring(0, 2) + notate(getSqPos(mv.substring(0, 2)) + 1)) != -1)
                            moves += ":" + mv; // not king-side castling through check
                    }
                }
            }
        }
        return moves;
    }
    
    /**
     * creates a concatenated String of possible moves, seperated by ":", for the player whose turn it is.  In addition, checks and act upon end-game situations.
     */
    protected void updateMoveList()
    {
        // get all the moves for the current player
        moveList = listMoves(((moveCount % 2 == 0) ? "w" : "b"), ALL_MOVES); // lists all the moves for current player
        // check for end-game situations
        if (moveList.length() == 0)
        { // no possible moves means one of two things (stalemate or checkmate)
            // check to see if the last move puts the opponent in check (puts an attack on the opponent's king)
            String lastMove = gameMoves[moveCount - 1]; // gets the last game move made
            boolean checked = (lastMove.indexOf("+") != -1); // gets the status of the attack on the opponent's king
            String pStr = ((!checked) ? "drawngame.png" : (moveCount % 2 == 0) ? "blackwins.png" : "whitewins.png"); // determines which image to display
            addObject(new Panel(pStr), 520, 300); // displays new panel image
            gameMoves[moveCount - 1] = gameMoves[moveCount - 1] + ((pStr.startsWith("draw")) ? " stalemate" : " checkmate"); // adds comment to move display
            info.updateImage(); // updates information panel (which displays game moves)
        }
    }
    
    // started() removes any panels that may be showing upon start-up
    /**
     * over-rides 'started()' in the World class to remove to title screen object, if present.
     */
    public void started()
    {
        removeObjects(getObjects(Panel.class));
    }
    
    // makeMove(String) takes a String representation of a move ("A1H8" format)
    // and executes the move internally (without displaying any changes)
    private void makeMove(String currentMove)
    {
        gameMoves[moveCount] = currentMove; // adds the move to execute to the game moves
        String startSq = currentMove.substring(0, 2), endSq = currentMove.substring(2, 4); // seperates move notation to individual square notation
        movers += (char) pieceInt[pieceAt(startSq) + 6]; // add the piece to the list of pieces that are moved for each move of the game
        // update the array holding the numbers of the moves that each castling possibility is permanently lost
        if (moveCastlingLost[0] == -1 && (currentMove.indexOf("e8") > -1 || currentMove.indexOf("a8") > -1)) moveCastlingLost[0] = moveCount;
        if (moveCastlingLost[1] == -1 && (currentMove.indexOf("e8") > -1 || currentMove.indexOf("h8") > -1)) moveCastlingLost[1] = moveCount;
        if (moveCastlingLost[2] == -1 && (currentMove.indexOf("e1") > -1 || currentMove.indexOf("a1") > -1)) moveCastlingLost[2] = moveCount;
        if (moveCastlingLost[3] == -1 && (currentMove.indexOf("e1") > -1 || currentMove.indexOf("h1") > -1)) moveCastlingLost[3] = moveCount;
        // record the move
        gameString += "" + getMoveChar(currentMove); // adds the move (and records changes in the moveCastlingLost array) to the condensed game move list
        moveCount++; // increments move counter
        adjustPosition(currentMove); // makes changes in the String 'position' to represent the position after the basic move given (currentMove)
        // deal with special moves
        if (currentMove.indexOf('<') != -1) adjustPosition((moveCount % 2 == 0) ? "a8d8" : "a1d1"); // moves rook in queen-side castling
        if (currentMove.indexOf('>') != -1) adjustPosition((moveCount % 2 == 0) ? "h8f8" : "h1f1"); // moves rook in king-side castling
        if (currentMove.indexOf('@') != -1)
            adjustPosition(getSqPos(startSq) + ((getSqPos(endSq) % 8) - (getSqPos(startSq) % 8)), " "); // removes captured pawn in en passant
        int pIndex = currentMove.indexOf('^');
        if (pIndex != -1) adjustPosition(getSqPos(endSq), currentMove.substring(pIndex + 1, pIndex + 2)); // changes piece type in pawn promotion
        // determine if opponent is put in check with the move
        String king = (moveCount % 2 == 0) ? "K" : "k"; // get String representation of king to look for
        boolean oppCheck = listMoves(((moveCount % 2 == 0) ? "b" : "w"), CAPTURES_ONLY).indexOf("x" + king) > -1; // determine if king is in check
        if (oppCheck) gameMoves[moveCount - 1] += "+"; // if opponent is checked, adds '+' to move string
    }
    
    // getMoveChar(String) takes a String representation of a move and converts it, along with any changes to moveCastlingLost,
    // into a single coded character, and return it
    private char getMoveChar(String currentMove)
    {
        int charVal = 0; // initializes local variable to return
        for (int i = 0; i < 4; i++) charVal = charVal * 2 + (moveCastlingLost[3 - i] == moveCount ? 1 : 0); // codes the changes in 'moveCastlingLost'
        for (int i = 0; i < 4; i++) charVal = charVal * 8 + ("abcdefgh12345678".indexOf(currentMove.charAt(3 - i)) % 8); // codes 'currentMove'
        return (char) charVal; // returns the coded data in character form
    }
        
    /**
     *  Searches and extracts the given move (from 'fromSq' to 'toSq') from the possible moves list and executes the move internally and externally. 
     *  In addition, will add promotion icons to the world, if neccessary.
     *  @param fromSq a string representation of the square a piece moves from (in "A1" format)
     *  @param toSq a string representation of the square the piece moves to (in "A1" format)
     */
    protected void makeMove(String fromSq, String toSq)
    {
        int mIndex = moveList.indexOf(":" + fromSq + toSq); // gets index of the move in 'moveList'
        String mv = moveList.substring(mIndex + 1, (moveList + ":").substring(mIndex + 1).indexOf(':') + mIndex + 1); // extracts the move from 'moveList'
        removeObjects(getObjects(Icon.class)); // removes any Icon objects currently displaying
        if (mv.indexOf('^') != -1) for (int i = 0; i < 4; i++) addObject(new Icon(getSqPos(mv.substring(2, 4)), i), 0, 0); // if pawn promotion, add new Icon objects
        makeMove(mv); // change 'position' to represent position after move
        updatePosition(); // display new board position
    }
    
    // takeBack(String) takes a String representation of a move ("A1H8" format)
    // and reverses the move internally (without displaying the changes)
    /**
     * takes the given 'move' and restores, internally, the state of the game prior to the move being internally executed.
     *
     * @param move a String representation of a move (in "A1H8" format)
     */
    protected void takeBack(String move)
    {
        // deal with special moves
        if (move.indexOf('>') != -1) adjustPosition((moveCount % 2 == 0) ? "f8h8" : "f1h1"); // moves rook back in king-side castling
        if (move.indexOf('<') != -1) adjustPosition((moveCount % 2 == 0) ? "d8a8" : "d1a1"); // moves rook back in queen-side castling
        String startSq = move.substring(0, 2), endSq = move.substring(2, 4); // seperates move notation into its two individual square notations
        if (move.indexOf('^') != -1) adjustPosition(getSqPos(endSq), (moveCount % 2 == 0) ? "p" : "P"); // changes piece back to pawn in pawn promotion
        // move the piece back from where it came
        adjustPosition(endSq + startSq); // returns moving piece back to starting square
        int xIndex = move.indexOf('x'); // gets index of capture symbol
        String pceTaken = (xIndex != -1) ? move.substring(xIndex + 1, xIndex + 2) : " "; // gets String representation of captured piece
        if (move.indexOf('@') != -1) pceTaken = (moveCount % 2 == 0) ? "P" : "p"; // if en passant, set String representation to opposite color pawn
        int pcePos = (move.indexOf('@') != -1) ? getSqPos(startSq) + ((getSqPos(endSq) % 8) - (getSqPos(startSq) % 8)) : getSqPos(endSq); // determines where piece is replaced
        adjustPosition(pcePos, pceTaken); // returns captured piece back to board
        moveCount--; // decrements move counter
        // update the array holding the number of the moves that each castling possibility is permanently lost
        if (moveCastlingLost[3] == moveCount) moveCastlingLost[3] = -1;
        if (moveCastlingLost[2] == moveCount) moveCastlingLost[2] = -1;
        if (moveCastlingLost[1] == moveCount) moveCastlingLost[1] = -1;
        if (moveCastlingLost[0] == moveCount) moveCastlingLost[0] = -1;
        movers = movers.substring(0, movers.length() - 1); // removes the move from the list of condensed moves
    }
    
    // updatePosition() displays the current position and moves played list, 
    // and prepares for the next move by creating a new possible move list
    /**
     * does finalization of moves by showing the new position, updating the display of game moves, and creating a new list of possible moves for the current player.
     */
    protected void updatePosition()
    {
        addPieces(); // shows the position represented by the String variable 'position'
        info.updateImage(); // shows the current list of moves played
        updateMoveList(); // create a new list of possible moves available to the current player
    }
    
    // notateSquare(int) takes a square number (0 to 63)
    // and returns a String representation of that square ("A1" format)
    private String notateSquare(int sqNum)
    {
        int x = sqNum % 8, y = (sqNum - x) / 8; // gets the coordinates of the square
        return ("abcdefgh".substring(x, x + 1) + "87654321".substring(y, y + 1)); // returns the String representation of the square
    }
    
    // adjustPosition(String) takes a String representation of a move ("A1H8" format)
    // and adjusts the String 'position' to reflect the position of the board after that move
    // (this only sets the 'from square' to empty and the 'to square' to what was on the 'from square',
    // additional changes may be required to complete the move)
    private void adjustPosition(String move)
    {
        String startSq = move.substring(0, 2), endSq = move.substring(2, 4); // seperates move notation to its two individual square notations
        adjustPosition(getSqPos(endSq), getOccupant(startSq)); // changes end square to what start square is
        adjustPosition(getSqPos(startSq), " "); // emptys start square
    }
    
    // adjustPosition(int, String) takes a square number and a String representation of a piece
    // and sets the specified character (square number) in the String 'position' to that piece
    private void adjustPosition(int sqNum, String piece)
    {
        String myAft = position.substring(sqNum + 1); // holds the hind part of the String
        String myFore = (sqNum > 0) ? position.substring(0, sqNum) : ""; // holds the fore part of the String
        position = myFore + piece + myAft; // changes the specific character (at 'sqNum') to 'piece'
    }
    
    /** 
     * Changes the pawn on the given square (sqNum), to the specified chessman (piece). NOTE: 
     * called by Icon.class when Icon object is clicked on.
     * @param sqNum the number of the square, in the range 0 to 63, that a pawn is promoting on
     * @param piece the type of piece that the pawn is promoting to
     */
    protected void changePromotion(int sqNum, String piece)
    {
        adjustPosition(sqNum, piece); // changes 'position' to show change in piece
        String mv = gameMoves[moveCount - 1]; // gets the last game move played
        int pIndex = mv.indexOf('^'); // gets the index of the pawn promotion symbol in the last move
        gameMoves[moveCount - 1] = mv.substring(0, pIndex + 1) + piece + mv.substring(pIndex + 2); // adjusts the piece shown to be promoted to in the last move
        updatePosition(); // displays the current position represented in the String 'position'
    }
    
    // getSqPos(String) takes a square notation ("A1" format) and converts it to a square number (0 to 63)
    // and returns that square number
    private int getSqPos(String sq) { return "abcdefgh".indexOf(sq.charAt(0)) + 8 * (7 - "12345678".indexOf(sq.charAt(1))); }    
    
    // getOccupant(int) takes a square number (0 - 63)
    // and returns a String representation of the piece residing on that square
    private String getOccupant(int sqPos) { return position.substring(sqPos, sqPos + 1); }    
    
    // getOccupant(String) takes a square notation ("A1" format)
    // and return a String representation of the piece residing on that square
    // { "k", "q", "b", "n", "p", " ", "P", "N", "B", "R", "Q", "K" } 
    // uppercase = white piece, lowercase = black piece, " " = empty square
    private String getOccupant(String square) { return getOccupant(getSqPos(square)); }
    
    // getMoves(int, int, boolean) takes a piece number, and square number and a flag to determine if all moves or just captures are to be listed
    // and return a listing of moves for that piece in a concatenated String (moves are separated by ":")
    private String getMoves(int pce, int sqNum, boolean listAll)
    {
        String pMvs = ""; // initializes local String to build the list of possible moves to return
        int x = sqNum % 8, y = (sqNum - x) / 8; // converts the square number ('sqNum') to its individual numeric coordinates
        int x2, y2, dx, dy, hold; // initializes local variables
        // branch depending on piece type (regardless of color)
        switch (pce * pce)
        {
        case 1: // PAWN [ pce = { 1 (WP), or -1 (BP) } ]
            // check pawn pushes (non-capture moves; when moving directly forward)
            y2 = y - pce; // pawn push one square forward
            if (listAll && pieceAt(x, y2) == 0)
            { // listing all moves and moving to empty square
                pMvs += ":" + notate(x, y) + notate(x, y2); // adds move to list of possibles
                if (7 - y2 * 2 == 7 * pce) pMvs = pMvs + "^" + ((pce == -1) ? "q" : "Q"); // if pawn promoting, complete move promoting to queen as default
                y2 = y - 2 * pce; // pawn push two squares forward
                if (listAll && y * 2 - 7 == 5 * pce && pieceAt(x, y2) == 0)
                { // listing all moves, pawn in position to move 2 squares and moving to empty square
                    pMvs += ":" + notate(x, y) + notate(x, y2); // adds move to list of possibles
                }
            }
            // check pawn moves that captures an opponent's piece
            for (int myCt = 0; myCt < 2; myCt++)
            { // for each capturing direction
                dx = 1 - 2 * myCt; // sets capturing direction
                x2 = x + dx; y2 = y - pce; // move pawn one square forward and toward to capturing direction
                if (onBoard(x2, y2) && pieceAt(x2, y2) * pce < 0)
                { // square valid and occupied by opposite color piece
                    pMvs += ":" + notate(x, y) + notate(x2, y2) + "x" + getOccupant(squareIndex(x2, y2)); // adds move to list of possibles
                    if (listAll && 7 - y2 * 2 == 7 * pce) pMvs += "^" + ((pce == -1) ? "q" : "Q"); // if pawn promoting, complete move promoting to queen as default
                }
            }
            // check for en-passant
            if (listAll && 7 - y * 2 == pce)
            { // listing all moves and pawn in position to capture en-passant
                for (int myCt = 0; myCt < 2; myCt++)
                { // for each capturing direction
                    dx = 1 - 2 * myCt; // sets capturing direction
                    x2 = x + dx; y2 = y - pce; // pawn move (en-passant) one square forward and one square in the capturing direction
                    if (onBoard(x2, y2))
                    { // square valid
                        String testMove = notate(x2, y2 - pce) + notate(x2, y); // code what last move should be if this were a valid en-passant move
                        if (gameMoves[moveCount - 1].substring(0, 4).equals(testMove) && "Pp".indexOf(position.charAt(squareIndex(x2, y))) > -1)
                        { // last move was of correct movement and by a pawn
                            pMvs = pMvs + ":" + notate(x, y) + notate(x2, y2) + "x" + getOccupant(squareIndex(x2, y2)) + "@"; // adds move to list of possibles
                        }
                    }
                }
            }
            break;
        case 4: // KNIGHT [ pce = { 2 (WN), or -2 (BN) } ]
            dx = 1; dy = 2; // initialize offsets
            for (int myCt = 0; myCt < 8; myCt++)
            { // for each direction
                x2 = x + dx; y2 = y + dy; // knight move to specified square
                if (onBoard(x2, y2) && pieceAt(x2, y2) * pce < 1 && (listAll || pieceAt(x2, y2) * pce < 0))
                { // square valid and not occupied by same color piece and either listing all moves or capture move
                    pMvs += ":" + notate(x, y) + notate(x2, y2) + ((pieceAt(x2, y2) * pce < 0) ? "x" + getOccupant(squareIndex(x2, y2)) : ""); // adds move to list of possibles
                }
                // change offset(s) for next loop
                dy = -dy;
                if (myCt % 2 == 1) dx = -dx;
                if (myCt == 3) { hold = dx; dx = dy; dy = hold; }
            }
            break;
        case 9: // BISHOP [ pce = { 3 (WB), or -3 (BB) } ]
            pMvs += getDiagonalMoves(pce, x, y, listAll); // adds all bishop moves to list of possibles
            break;
        case 16: // ROOK [ pce = { 4 (WR), or -4 (BR) } ]
            pMvs += getAxialMoves(pce, x, y, listAll); // adds all rook moves to list of possibles
            break;
        case 25: // QUEEN [ pce = { 5 (WQ), or -5 (BQ) } ]
            pMvs += getDiagonalMoves(pce, x, y, listAll) + getAxialMoves(pce, x, y, listAll); // adds all queen moves to list of possibles
            break;
        case 36: // KING [ pce = { 6 (WK), or -6 (BK) } ]
            dx = 0; dy = -1; // initialize offsets for standard king moves
            int ddx = 1, ddy = 1; // initialize local directionals
            for (int myCt = 0; myCt < 8; myCt++)
            { // for each square surrounding the king
                x2 = x + dx; y2 = y + dy; // set coordinates of square to move to
                if (onBoard(x2, y2) && pieceAt(x2, y2) * pce < 1 && (listAll || pieceAt(x2, y2) * pce < 0))
                { // square valid and not occupied by same color piece and either listing all moves or capture move
                    String capStr = (pieceAt(x2, y2) * pce < 0) ? "x" + getOccupant(squareIndex(x2, y2)) : ""; // create String representation of capture
                    pMvs += ":" + notate(x, y) + notate(x2, y2) + capStr; // adds move to list of possibles
                }
                // adjust directionals and offsets for next square
                if (dx == 0) dx += ddx; // if in middle of moving left or right, keep moving that way
                else if (dy == 0) dy += ddy; // if in middle of movin up or down, keep moving that way
                else if (ddx == dx) { ddx = -ddx; dy += ddy; } // if done moving left or right, change directional and start moving up or down
                else if (ddy == dy) { ddy = -ddy; dx += ddx; } // if done moving up or down, change directional and start moving left or right
            }
            // check for castling moves
            if (listAll) // castling moves
            { // listing all moves (which would include castling moves)
                int castlePos = 2 * ((moveCount + 1) % 2); // get starting index in 'moveCastlingLost' depending on color of piece
                // check queen side castling
                if (moveCastlingLost[castlePos] == -1)
                { // queen side castling has yet to be permanently removed as a possibility
                    boolean test1 = " ".equals(getOccupant(sqNum - 1)); // get occupation of queens square
                    boolean test2 = " ".equals(getOccupant(sqNum - 2)); // get occupation of queens-bishop square
                    boolean test3 = " ".equals(getOccupant(sqNum - 3)); // get occupation of queens-knight square
                    if (test1 && test2 && test3) 
                    { // all squares between king and queen's rook are empty
                        pMvs = pMvs + ":" + notate(x, y) + notate(x - 2, y) + "<"; // adds move to list of possibles
                    }
                }
                // check king side castling
                if (moveCastlingLost[castlePos + 1] == -1)
                { // king side castling has yet to be permanently removed as a possibility
                    boolean test1 = " ".equals(getOccupant(sqNum + 1)); // get occupation of kings-bishop square
                    boolean test2 = " ".equals(getOccupant(sqNum + 2)); // get occupation of kings-knight square
                    if (test1 && test2) 
                    { // both squares between king and king's rook are empty
                        pMvs = pMvs + ":" + notate(x, y) + notate(x + 2, y) + ">"; // adds move to list of possibles
                    }
                }
            }
            break;
        }
        return pMvs; // returns the requested list of possible moves
    }
    
    // getDiagonalMoves(int, int, int, boolean) takes an integer piece type, column (0 to 7) and row (0 to 7) numbers, and a flag
    // to determine whether all moves are listed or just the capture moves (for finding attacks on kings: checks)
    // and builds a listing of moves along the four diagonals from the square (x, y) (for a queen or bishop).  The moves are listed
    // in a concatenated String which is returned to the calling method (moves are separated by ":").
    private String getDiagonalMoves(int pceType, int x, int y, boolean listAll)
    {
        int dx = 1, dy = 1; // initialize directionals
        int x2, y2; // initialize coordinates of square to move to
        String bMvs = ""; // initialize local String to build the list of possible moves to return
        for (int myCt = 0; myCt < 4; myCt++)
        { // for each diagonal radiating from current location (x, y)
            int dd = 0; // initialize distance
            while (true)
            {
                dd++; // increment distance 
                x2 = x + dd * dx; y2 = y + dd * dy; // adjust coordinate of square to move to
                if (!onBoard(x2, y2)) break; // square invalid, end processing this direction
                int onType = pieceAt(x2, y2); // gets integer representation of piece occupying square to move to
                if (onType * pceType < 1 && (listAll || onType != 0))
                { // not occupied by piece of same color and either listing all moves or capture move
                    String capStr = (onType != 0) ? "x" + getOccupant(squareIndex(x2, y2)) : ""; // creates String representation of capture
                    bMvs = bMvs + ":" + notate(x, y) + notate(x2, y2) + capStr; // adds move to list of possibles
                }
                if (onType != 0) break; // if piece occupying square to move to, end processing this direction
            }
            // adjust directionals for next direction
            dy = -dy;
            if (myCt % 2 == 1) dx = -dx;
        }
        return bMvs; // returns list of possible diagonal moves
    }
    
    // getAxialMoves(int, int, int, boolean) takes an integer piece type, column (0 to 7) and row (0 to 7) numbers, and a flag
    // to determine whether all moves are listed or just the capture moves (for finding attacks on kings = checks)
    // and builds a listing of moves along the horizontal and vertical from the square (x, y) (for a queen or rook).  The moves are listed
    // in a concatenated String which is returned to the calling method (moves are separated by ":").
    private String getAxialMoves(int pce, int x, int y, boolean listAll)
    {
        int dx = 0, dy = 1; // initialize directionals
        int x2,  y2; // initialize coordiinates of square to move to and
        String rMvs = ""; // initialize local String to build the list of possible moves to return
        for (int myCt = 0; myCt < 4; myCt++)
        { // for each axial direction radiating from current location (x, y)
            int dd = 0; // initialize distance
            while (true)
            {
                dd++; // increment distance
                x2 = x + dd * dx; y2 = y + dd * dy; // adjust coordinates of square to move to
                if (!onBoard(x2, y2)) break; // if square invalid, end processing this direction
                int onType = pieceAt(x2, y2); // gets integer representation of piece occupying square to move to
                if (onType * pce < 1 && (listAll || onType != 0))
                { // not occupied by same color piece and either listing all moves or capture move
                    String capStr = (onType != 0) ? "x" + getOccupant(squareIndex(x2, y2)) : ""; // gets String representation of capture
                    rMvs = rMvs + ":" + notate(x, y) + notate(x2, y2) + capStr; // adds move to list of possibles
                }
                if (onType != 0) break; // if piece occupying square to move to, end processing this direction
            }
            // adjust directionals for next direction
            int hold = dx; dx = dy; dy = hold;
            if (myCt % 2 == 1) dy = -dy;
        }
        return rMvs; // returns list of possible axial moves
    }
    
    // notate(int) takes a square number (0 to 63)
    // and returns a String representing of square ("A1" format)
    private String notate(int sq)
    {
        int x = sq % 8, y = (sq - x) / 8;
        return notate(x, y);
    }
    
    // notate(int, int) take a square's column (0 to 7, across) and row (0 to 7, down)
    // and returns a String representation of that square ("A1" format)
    private String notate(int x, int y) { return "abcdefgh".substring(x, x + 1) + "87654321".substring(y, y + 1); }
        
    // onBoard(int, int) takes a column (0 to 7, hopefully) number and row (0 to 7, hopefully) numbers
    // and returns true if they land on the board (both between 0 and 7 inclusive)
    private boolean onBoard(int x, int y) { return x > -1 && x < 8 && y > -1 && y < 8; }
    
    // pieceAt(int) takes a square number
    // and returns the number of the type of piece on that square (-6 to 6)
    private int pieceAt(int sq) { return "kqrbnp PNBRQK".indexOf(position.charAt(sq)) - 6; }
    
    private int pieceAt(String coord)
    {
        int x = "abcdefgh".indexOf(coord.charAt(0));
        int y = "87654321".indexOf(coord.charAt(1));
        return pieceAt(x, y);
    }
    
    // pieceAt(int, int) takes a square's column (0 to 7) and row (0 to 7) numbers
    // and returns the number of the type of piece on that square (-6 to 6)
    private int pieceAt(int x, int y) { return pieceAt(squareIndex(x, y)); }
    
    // squareIndex(int, int) takes a square's row (0 to 7) and column (0 to 7) numbers and coverts them to a square number
    // and returns that square number (0 to 63)
    private int squareIndex(int x, int y) { return y * 8 + x; }
    
    // // printMovesToTerminal() print the moves played in the game in the terminal window
    // // (called from immediate mode only; while scenario is NOT running, right-click in world to select and run this method)
    // /**
     // * prints out the game moves to the output terminal. There is no internal call for this.  
     // * Stop the execution, right-click on the world (any empty square) and select this method to execute it.
     // */
    // public void printMovesToTerminal()
    // {
        // String output = "MV |  WHITE      |  BLACK\n---+-------------+--------------\n";
        // String tab = "                              ";
        // for (int i = 0; i < (moveCount + 1) / 2; i++)
        // {
            // String line = " " + (i+1)+" |  ";
            // line = line.substring(line.length()-6, line.length());
            // line += gameMoves[i * 2] + tab;
            // line = line.substring(0, 17) + "|  ";
            // if (moveCount > i * 2 + 1)  line += gameMoves[i * 2 + 1];
            // output += line+"\n";
        // }
        // JTextArea text = new JTextArea(output);
        // JOptionPane.showMessageDialog(null, text, "Copy and paste as standard text", JOptionPane.INFORMATION_MESSAGE);
    // }
    
    // public void printMovesListToTerminal()
    // {
        // String output = "[ ";
        // String[] sqsPce = new String[64];
        // for(int i=0; i<64; i++) sqsPce[i]=""+"rnbqkbnrpppppppp                                PPPPPPPPRNBQKBNR".charAt(i);
        // for(int i=0; i<moveCount; i++)
        // {
            // String m = gameMoves[i];
            // String f1 = ""+m.charAt(0);
            // String r1 = ""+m.charAt(1);
            // String f2 = ""+m.charAt(2);
            // String r2 = ""+m.charAt(3);
            // String s1 = " "+(f1+r1).toUpperCase();
            // int fr = "87654321".indexOf(r1)*8+"abcdefgh".indexOf(f1);
            // int to = "87654321".indexOf(r2)*8+"abcdefgh".indexOf(f2);
            // String p = sqsPce[fr];
            // if (m.indexOf('^') >= 0) p = ""+m.charAt(m.indexOf('^')+1);
            // String s2=p+(f2+r2).toUpperCase();
            // sqsPce[fr] = " ";
            // sqsPce[to] = p;
            // String s3 = "";
            // String s4 = "";
            // if (m.indexOf('@') >= 0) s3 = " "+(f2+r1).toUpperCase();
            // if (m.indexOf('>') >= 0 || m.indexOf('<') >= 0)
            // {
                // s3 = " "+("c".equals(f2) ? "A" : "H")+r2;
                // s4 = ("1".equals(r2) ? "R" : "r")+("c".equals(f2) ? "D" : "F")+r2;
            // }
            // if (!"".equals(s3))
            // {
                // p = ""+s3.charAt(0);
                // f1 = ""+s3.charAt(1);
                // r1 = ""+s3.charAt(2);
                // fr = "87654321".indexOf(r1)*8+"ABCDEFGH".indexOf(f1);
                // sqsPce[fr]=p;
            // }
            // if (!"".equals(s4))
            // {
                // p = ""+s4.charAt(0);
                // f2 = ""+s4.charAt(1);
                // r2 = ""+s4.charAt(2);
                // to = "87654321".indexOf(r2)*8+"ABCDEFGH".indexOf(f2);
                // sqsPce[to]=p;
            // }
            // output += "'"+s1+s2+s3+s4+"'";
            // if (i+1<moveCount) output += ", ";
            // if ((i+1)%10 == 0 && i+1<moveCount) output += "\n";
        // }
        // output += " ],\n";
        // JTextArea text = new JTextArea(output);
        // JOptionPane.showMessageDialog(null, text, "Copy and paste to HTML javascript code", JOptionPane.INFORMATION_MESSAGE);
        
    // }
    
    // public void act()
    // {
        // String key = Greenfoot.getKey();
        // if ("p".equals(key)) printMovesToTerminal();
        // if ("l".equals(key)) printMovesListToTerminal();
    // }
}
